from . import jacklib
from queue import Queue, Empty as QuequeEmpty

# Globals
global jack_client, jack_midi_in_port, jack_midi_out_port, jack_midi_in_data, jack_midi_out_data
global jack_midi_in_is_open
jack_midi_in_is_open = True #A switch to 'mute' midi in for Laborejo without destroying any connections. The midi in data is just discarded.
jack_client = None
jack_midi_in_port  = None
jack_midi_out_port = None
jack_midi_in_data  = Queue(512)
jack_midi_out_data = Queue(512)

static_event = jacklib.jack_midi_event_t()
static_mtype = jacklib.c_ubyte*3

class States(object):
    """Change the property functions to switch the midi controller"""
    def __init__(self):
        """Midi in modes"""
        self.pitchbend = 64 #middle
        self.cc = {
                    0x07:127, #Volume is high
                    0x40:0, #Hold/Sustain Pedal is off.
                    }

    def digitalCC(self, cc):
        if cc in self.cc:
            if self.cc[cc] >= 64:
                return True
            else:
                return False
        else:
            return False

states = States()

def jack_process_callback(nframes, arg):
    global jack_midi_in_port, jack_midi_out_port, jack_midi_in_data, jack_midi_out_data

    # MIDI In
    midi_in_buffer = jacklib.port_get_buffer(jack_midi_in_port, nframes)

    if (midi_in_buffer):
      event_count = jacklib.midi_get_event_count(midi_in_buffer)

      for i in range(event_count):
        if (jacklib.midi_event_get(jacklib.pointer(static_event), midi_in_buffer, i) == 0):
          if (static_event.size == 1):
            jack_midi_in_data.put_nowait((static_event.buffer[0], 0, 0))
          elif (static_event.size == 2):
            jack_midi_in_data.put_nowait((static_event.buffer[0], static_event.buffer[1], 0))
          elif (static_event.size >= 3):
            jack_midi_in_data.put_nowait((static_event.buffer[0], static_event.buffer[1], static_event.buffer[2]))

          if (jack_midi_in_data.full()):
            break

    # MIDI Out
    midi_out_buffer = jacklib.port_get_buffer(jack_midi_out_port, nframes)

    if (midi_out_buffer):
      jacklib.midi_clear_buffer(midi_out_buffer)

      if (jack_midi_out_data.empty() == False):
        while (True):
          try:
            mode, note, velo = jack_midi_out_data.get_nowait()
          except QuequeEmpty:
            break

          data = static_mtype(mode, note, velo)
          jacklib.midi_event_write(midi_out_buffer, 0, data, 3)

          jack_midi_out_data.task_done()


    return 0

#These are here to be overwritten.
def noteOn(arg):
    pass
def restOn():
    pass
def addChordNote():
    pass
def pitchbend(x):
    pass

smfChannelCache = 0 #the current tracks smf channel.
executeCC = {} #populate with key(controller number):function. If there is any controller change.

def midiInChooser():
    """This is the function that is called by the program periodically.
    The API/Core do not activate it. The user program must start this
    and keep it running. All GUIs are considered user programs as well.
    Laborejos Qt GUI uses a Qt timer to keep midiInChooser running
    """
    global jack_midi_in_data, jack_midi_out_data
    global jack_midi_in_is_open

    if (jack_midi_in_data.empty() == False):
        while (True):
            try:
                data1, data2, data3 = jack_midi_in_data.get_nowait()
            except QuequeEmpty:
                break

            channel = (data1 & 0x0F)#+1
            mode    =  data1 & 0xF0

            #Midi Thru. The listener. Has nothing to do with note entry. It can be used to pre-playback sounds to your sampler.
            #Sends only note on and off.
            if mode == 0x90: #on
                jack_midi_out_data.put_nowait((mode + smfChannelCache, data2, data3)) #create a note on the currents track channel.
            elif mode == 0x80: #off
                jack_midi_out_data.put_nowait((mode + smfChannelCache, data2, data3))

            if mode == 0x90 and jack_midi_in_is_open:  #the midi in port can be muted: #note on. Note off gets ignored.
                if states.digitalCC(0x07):
                    if states.digitalCC(0x40):
                        addChordNote(data2)
                    else:
                        noteOn(data2)
                else:
                    restOn()
            elif mode == 0xB0:  #CC
                states.cc[data2] = data3 #save the controller data
                if data2 in executeCC:
                    executeCC[data2]() #execute the bound function, if existant.

            jack_midi_in_data.task_done()



if jacklib.jacklib:
    # Start jack
    jack_status = jacklib.jack_status_t(0)
    jack_client = jacklib.client_open("Laborejo", jacklib.JackNoStartServer, jacklib.pointer(jack_status))
    #jack_client = jacklib.client_open("Laborejo", jacklib.JackNoStartServer, 0)
    if jack_client:
        jack_midi_in_port = jacklib.port_register(jack_client, "single_step_input", jacklib.JACK_DEFAULT_MIDI_TYPE, jacklib.JackPortIsInput, 0)
        jack_midi_out_port = jacklib.port_register(jack_client, "midi_thru", jacklib.JACK_DEFAULT_MIDI_TYPE, jacklib.JackPortIsOutput, 0)
        jacklib.set_process_callback(jack_client, jack_process_callback, None)
    else:
        print ("warning: Jack Server not running. Starting without JACK.")
        jack_client = None
else:
    print ("warning: libjack not found. Laborejo JACK client broken.")
    jack_client = None

def start():
    jacklib.activate(jack_client)
    #midiIn.start() #Let a higher level, like a GUI, decide.

# Close Jack
def close():
    if (jack_client):
        jacklib.deactivate(jack_client)
        jacklib.client_close(jack_client)
