# -*- coding: utf-8 -*-

import sys
import subprocess, operator, os, warnings, inspect
from random import random, randint, shuffle
from collections import defaultdict
localepath = os.path.join(os.path.dirname(__file__), 'locale')

from PyQt4 import QtCore, QtGui
def customTranslate(one, theString, three, four):
    #self.actionForte.setText(QtGui.QApplication.translate("MainWindow", "𝆑 forte", None, QtGui.QApplication.UnicodeUTF8))
    x =  __translateOrg(one, theString, three, four)
    return x.strip("\x00")  #there are \x00 in the strings which need to removed before display. This is most likely a pyqt/Python3 bug

__translateOrg = QtGui.QApplication.translate
QtGui.QApplication.translate = customTranslate #replace Qt function with our own "translate".

import laborejocore as api #This starts the backend and automatically creates a session. But not a score/movement/tracks.

#Start the Gui
from laborejoqt.designer.gui_main import Ui_MainWindow
from laborejoqt.designer.gui_musictab import Ui_musicTab
#from designer.gui_megaman import Ui_Form
app = QtGui.QApplication(sys.argv) #Start qt as early as possible so that item creation on program startup works. Otherwise the instances create segfaults because qt is not ready to create QGraphicsTextItem (in items.py)
if QtCore.QLocale().languageToString(QtCore.QLocale().language()) == "German":
    translator = QtCore.QTranslator()
    translator.load("trans_de.qm", "laborejoqt/designer") #German
    app.installTranslator(translator)

from laborejoqt import config, dictionaries, modalcommands, listener
from laborejoqt import items as guiitems
from laborejoqt.constants import *
from laborejoqt.container import GuiTrack, PSEUDO
from laborejoqt.dynformcreator import fedit #Generate dialogs/formulars with a simple syntax.

class StartQT4(QtGui.QMainWindow):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        config.reloadPersonalConfig()  #override global qt-config data with user data.
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.statusBar().showMessage('Ready')
        self.ui.Tab.clear()
        self.session = Session()  #This is all the data.
        self.more = ""  #used in the status bar.
        self.mode = "No Mode"
        self.genericKeyActions = []  # All actions that have a shortcut like Return or Backspace. It filled automatically during shortcut creation
        self.shortcutDict = {} #for all static shortcuts.
        self.hoverShortcutDict = {} #for temporary hover shortcuts
        self.lastHover = None #the last time the mouse hovered over a menu it was this Qt.Action.
        self.recentFilesMenuActions = [] #temp storage. Only menu actions.
        self.allActions = [] #A quick access point for all actions.
        self.colors = [] #All used TabColors. this is never cleared. Who cares, there are enough colors.

        #Last Known Dirs. Remember for convenience
        self.last_open_dir = HOME
        self.last_save_dir = HOME
        self.last_export_dir = HOME
        self.last_import_dir = HOME
        self.last_subst_dir = HOME

        #A lilypond window that can be closed with Escape
        self.ui.lilyfloat = QtGui.QTextBrowser()  #somehow it must be part of main.ui, no matter the name. Is hidden at first.
        lilyEscapeAction = QtGui.QAction(QtGui.QIcon(), '', self)
        lilyEscapeAction.setShortcut("Escape")
        self.ui.lilyfloat.addAction(lilyEscapeAction)
        QtCore.QObject.connect(lilyEscapeAction,QtCore.SIGNAL("triggered()"), self.ui.lilyfloat.hide)


        #small terminal
        self.ui.smallTerminal.hide()
        self.ui.actionSmall_Terminal.setChecked(False)
        try:
            self.ui.terminal.setPlainText(open(api._getConfigDir() + 'laborejo-qt-temp-terminal', "r").read())
        except:
            pass

        def terminalKeypress(event):
            if event.modifiers() == QtCore.Qt.ControlModifier and event.key() == 16777220: #Ctrl Return
                self.processSmallTerminalInput()
            else:
                terminalOriginalKeypress(event)

        terminalOriginalKeypress = self.ui.terminal.keyPressEvent
        self.ui.terminal.keyPressEvent = terminalKeypress


        #Tabs
        ######
        self.ui.Tab.tabCloseRequested.connect(self.closeTabWarning)
        self.ui.Tab.currentChanged.connect(self.tabChangeEvent)

        #Menus
        ######
        #Remember the last hovered menu action
        def temphover(action):
            self.lastHover = action
        self.ui.menubar.hovered.connect(temphover)
        self.ui.menubar.keyPressEvent = self.menuKeyPressInterceptor

        qMenumousePressEvent = QtGui.QMenu.mousePressEvent

        def menuMiddleMouseButtonInterceptor(menuobject, event):
            #main already exists when this is used by the mouse.
            if event.button() == QtCore.Qt.MiddleButton:
                self.ui.toolBar.addAction(self.lastHover)
            else:
                qMenumousePressEvent(menuobject, event)

        def toolbarMiddleMouseButtonInterceptor(toolbarAction, event):
            if event.button() == QtCore.Qt.MiddleButton:
                self.ui.toolBar.removeAction(self.ui.toolBar.actionAt(event.pos()))

        QtGui.QMenu.mousePressEvent = menuMiddleMouseButtonInterceptor
        QtGui.QToolBar.mousePressEvent = toolbarMiddleMouseButtonInterceptor


        #Create User commands and menus
        self.generateUserCommands()
        self.generateSubstitutionsMenu()
        self.generateRecentFilesMenu()

        #Create Extra Toolbar Buttons
        self.generateToolbar()

        #Connect all preconfigured menu actions (not generated, not user) with backend functions
        for qtAction, function in dictionaries.menus.items():
            self.connectMenuAction(eval("self.ui." + qtAction) , eval(function))

        #Create All Shortcuts
        for key, value in config.shortcuts.items(): #key is really the keyboard key. value is the action name.
            if not "lambda" in value: #this should eleminate anything with real paths, like midi export.
                value = value.replace("/", "µ") #The user is allowed to give / as action names because it fits the user dir for user scripts in ~/.laborejo/scripts. But we can't use those as variable names, we use µ instead.
            if not "QtCore.Qt.KeypadModifier" in key:
                #for config: "QtCore.Qt.KeypadModifier + QtCore.Qt.Key_5" : "api.tail",
                key2 = key.title()
                keypad = False
            else:
                key2 = QtGui.QKeySequence(eval(key))
                keypad = True

            #Try to get the functiond/object. This is not an execution, just checking if it is there.
            try:
                #Non-menu actions.
                #These are mostly navigational keys like cursor,
                #home/end as well as some duplicate shortcuts like AddDot
                #which exists twice, once as menu and once on the Tab
                #key because it is near the note entry numbers

                #We need to create an action for these. Only actions can have shortcuts.
                eval(value)  #The action exists without the "self.ui." prefix.
                action = QtGui.QAction(QtGui.QIcon(), '', self)
                action.setShortcut(key2)
                self.addAction(action)
                if not keypad:
                    self.shortcutDict[key2] = action
                self.connectMenuAction(action, eval(value)) #despite the name this is not only for menu functons

            except NameError:
                #Menu Actions
                #The action already exists, we don't need to create a new one.
                action = eval("self.ui." + value)
                action.setShortcut(key2)
                if not keypad:
                    self.shortcutDict[key2] = action

            if key in ["Return", "Tab", "Backspace", "Shift+Tab", "Enter", "Left", "Right", "Up", "Down"]:
                self.genericKeyActions.append(action)


        #Movement Substitutions Empty Menu
        self.placeholderMovementSubstitutions = QtGui.QMenu()
        tmp = self.ui.menuAdvanced.addMenu(self.placeholderMovementSubstitutions)
        tmp = self.ui.menuAdvanced.insertMenu(self.ui.actionRemove_Substitution, self.placeholderMovementSubstitutions )
        tmp.setText('Movement Substitutions')

        #Movement Container Empty Menu
        self.placeholderMovementContainers = QtGui.QMenu()
        tmp2 = self.ui.menuContainer.insertMenu(self.ui.actionInsert_Empty_Container, self.placeholderMovementContainers )
        tmp2.setText('Movement Container')

        self.switchActionsOff()

        #Show the welcome tab
        #TODO
        #self.welcomeTab = WelcomeTab()
        #self.ui.Tab.addTab(self.welcomeTab, "Welcome")

        #Finally Show the Qt Window
        self.show()

    def processSmallTerminalInput(self):
        text = self.ui.terminal.toPlainText()
        f2 = open(api._getConfigDir() + 'laborejo-qt-temp-terminal', "w")
        f2.write(text)
        f2.close()
        if text:
            exec(text)

    def switchActionsOff(self):
        """Deactivate nearly all Menu Actions and Shortcuts
        as long as no file is open.
        (Except recent files which are not in this list on purpose)"""
        for a in self.allActions:
            a.setEnabled(False)
        allowed = [self.ui.actionNew, self.ui.actionOpen, self.ui.actionOpen_as_Template, self.ui.actionQuit, self.ui.actionChat, self.ui.actionManual, self.ui.actionAbout, self.ui.actionImportLisalo, self.ui.actionImportLisaloBus, self.ui.actionSmall_Terminal]
        for a in allowed:
            a.setEnabled(True)

    def switchActionsOn(self):
        for a in self.allActions:
            a.setEnabled(True)

    def toggleBarlines(self):
        """Switch GUI barlines on or off.
        This alone is only a relay station and is not a switch in itself
        toggleBarlines in main decides what to do"""
        if config.BARLINES:
            config.BARLINES = False
        else:
            config.BARLINES = True

        for backendWs, guiWs in self.session.scenes.items():
            guiWs.toggleBarlines(onOff = config.BARLINES)

        self.ui.actionBarlines.setChecked(config.BARLINES)

    def toggleVisible(self, qwidgetString, menu):
        qwidget = eval("self.session.currentGuiWorkspace.ui." + qwidgetString)

        #Toggle Off
        if qwidget.isVisible():
            if qwidgetString is "lyrics":  #delete all lyric items for the whole musicTab
                config.LYRICS = False
                for tr in self.session.currentGuiWorkspace.tracks:
                    tr._removeLyrics()
                self.session.currentGuiWorkspace.repositionTracks()
            qwidget.hide()
            menu.setChecked(False)
            self.session.currentGuiWorkspace.ui.score.setFocus(True)
            return False

        #Toggle On
        else:
            if qwidgetString is "lyrics":  #recreate all lyric items for the whole musicTab
                config.LYRICS = True
                for tr in self.session.currentGuiWorkspace.tracks:
                    tr.drawLyrics()
                self.session.currentGuiWorkspace.repositionTracks()
            qwidget.show()
            menu.setChecked(True)
            qwidget.setFocus(True)
            return True

    def toggleSmallTerminal(self):
        qwidget = self.ui.smallTerminal
        menu = self.ui.actionSmall_Terminal
        if qwidget.isVisible():
            qwidget.hide()
            menu.setChecked(False)
            self.focusOnActiveScore()
            return False
        else:
            qwidget.show()
            menu.setChecked(True)
            self.ui.terminal.setFocus(True)
            return True

    def focusOnActiveScore(self):
        self.session.currentGuiWorkspace.ui.score.setFocus(True)

    def zoomIn(self):
        """only a redirection to the current workspace.
        Without redirection we get no menu and shortcut"""
        self.session.currentGuiWorkspace.zoomIn()

    def zoomOut(self):
        """only a redirection to the current workspace.
        Without redirection we get no menu and shortcut"""
        self.session.currentGuiWorkspace.zoomOut()

    def widen(self):
        """only a redirection to the current workspace.
        Without redirection we get no menu and shortcut"""
        self.session.currentGuiWorkspace.widen()

    def shrinken(self):
        """only a redirection to the current workspace.
        Without redirection we get no menu and shortcut"""
        self.session.currentGuiWorkspace.shrinken()

    def tabChangeEvent(self):
        if api.core.session.workspace and isinstance(self.ui.Tab.currentWidget(), MusicTab):
            api.core.session.workspace = self.ui.Tab.currentWidget().data
            guiWs = main.session.scenes[api.core.session.workspace]
            self.ui.actionTrack_Properties.setChecked(guiWs.ui.trackProperties.isVisible())
            self.ui.actionMinimap.setChecked(guiWs.ui.minimap.isVisible())
            self.ui.actionLyrics.setChecked(guiWs.ui.lyrics.isVisible())
            guiWs.generateMovementSubstitutionsMenu()
            guiWs.generateMovementContainerMenu()
            guiWs.cursorPitch.pitch = api._getCursorPitch()
            guiWs.cursorPitch.update()
            if self.ui.lilyfloat.isVisible():
                self.ui.lilyfloat.setText(api.generateLilypond())

    def menuKeyPressInterceptor(self, qKeyEvent):
        """If you press a key while hovering with your mouse
        over the entry it sets a new shortcut for this item."""
        if qKeyEvent.nativeModifiers(): #0 or 1 for shift
            key = "Shift+" + qKeyEvent.text() #TODO: this fails on non-letter keys like < and >. the shift variant gets shift+the shifted char. Shift+> which is impossible.
        else: #no shift
            key = qKeyEvent.text()
        key = key.title()
        if not key in self.shortcutDict:
            if key in self.hoverShortcutDict: #clear the old shortcut first
                self.hoverShortcutDict[key].setShortcut(QtGui.QKeySequence())
            self.hoverShortcutDict[key] = self.lastHover
            self.lastHover.setShortcut(key)

    def connectMenuAction(self, action, function, append = True):
        """Connect the menu action with the backend function,
        or qt function in a few cases like Save and Open.
        The actions already exist in their places and
        no shortcuts are getting set here.

        Also update the GraphicsView/Viewport because we deactivated
        auto update. This prevents in-between movement of cursors,
        screenjumping and an over-all feeling of sickness :) """
        #self.session.currentGuiWorkspace.ui.score.setViewportUpdateMode(QtGui.QGraphicsView.MinimalViewportUpdate)
        #self.session.currentGuiWorkspace.ui.score.setViewportUpdateMode(QtGui.QGraphicsView.SmartViewportUpdate)
        #self.session.currentGuiWorkspace.ui.score.setViewportUpdateMode(QtGui.QGraphicsView.NoViewportUpdate)
        #self.session.currentGuiWorkspace.ui.score.setViewportUpdateMode(QtGui.QGraphicsView.FullViewportUpdate)
        if append:
            self.allActions.append(action)
        def funcWrap():
            function()
            try:
                self.session.currentGuiWorkspace.ui.score.viewport().update() #Update the grahipcsview manually
                self.session.currentGuiWorkspace.updateStatusBar()
            except:
                pass
        QtCore.QObject.connect(action, QtCore.SIGNAL("triggered()"), funcWrap)


    def uncheckModeActions(self):
        """Uncheck all menu items in menu 'Mode'"""
        for x in self.ui.menuMode.actions():
            x.setChecked(False)

    def defaultModalCommands(self):
        self.uncheckModeActions()
        self.ui.actionInsert_Duration.setChecked(True)
        self.mode = "Insert"
        modalcommands.commands.defaults()

    def prevailingMode(self):
        """Normal keys choose prevailing duration, alt mod inserts"""
        self.uncheckModeActions()
        self.ui.actionPrevailing_Duration.setChecked(True)
        self.mode = "Prevailing"
        modalcommands.commands.prevailing()

    def toggleMode(self):
        if self.mode == "Prevailing":
            self.defaultModalCommands() #Default number keys
        else:
            self.prevailingMode()


    def generateToolbar(self):
        """Read the toolbar dict from the config and user config file
        and convert it to toolbar buttons"""
        for key, value in config.toolbar.items():
            value = value.replace("/", "µ") #The user is allowed to give / as action names because it fits the user dir for user scripts in ~/.laborejo/scripts. But we can't use those as variable names, we use µ instead.
            #unlike shortcuts we only accept real menu actions here.
            self.ui.toolBar.addAction(eval("self.ui." + value))
            (eval("self.ui." + value)).setText(key.title())
            #self.allActions.append(eval("self.ui." + value))

    def addRecentFiles(self, fileName):
        """Add one file to the recent save file. If there are too much
        in that file delete the oldest entry"""
        configRecent = api._getConfigDir() + 'recent'

        try:
            f = open(configRecent)
        except:
            f = None
            f_data = ""
        if f:
            f_data = f.read()
            f.close

        recentList = f_data.split()

        if fileName in recentList:
            return False
        else:
            recentList.append(fileName)
            if len(recentList) >6:
                recentList.pop(0)

            finalString = "\n".join(recentList)

            f2 = open(configRecent, "w")
            f2.write(finalString)
            f2.close()
            self.generateRecentFilesMenu()

    def generateRecentFilesMenu(self):
        """Generate the recent files menu.
        This is the step after self.addRecentFiles.

        This always addes the complete recent files file to the menu."""

        #First remove the old files menu actions
        for menuAction in self.recentFilesMenuActions:
            self.ui.menuRecent_2.removeAction(menuAction)
        self.recentFilesMenuActions = []

        #Try to open the file where recent values are stored.
        try:
            f = open(api._getConfigDir() + 'recent', 'r')
        except IOError:
            f = None

        #Create a new recent files menu from the save file.
        if f:
            nameList = [fString for fString in f]
            nameList.reverse()
            for fString in nameList:
                fString = fString.rstrip()
                menuAction = self.ui.menuRecent_2.addAction(fString)
                self.recentFilesMenuActions.append(menuAction)
                self.connectMenuAction(menuAction, lambda fString = fString: self.load(directLoad = fString), append = False) #Don't append to the activate/deactivate list.
            f.close()

    def generateUserCommands(self):
        """Parse ~/.laborejo/scripts/ and create a menu structure
        which will execute the files in there.
        This is done by creating actions and BEFORE setting shortcuts.
        So the user can bind their own shortcuts in qt-lab.config.
        The name of the action is "user_" plus the name of the path
        without  ~/.laborejo/scripts/ and the tailing ".py"
        Any / is substituted with µ
        Example:
        ~/.laborejo/scripts/somedir/temp.py would be
        self.ui.user_somedirµtemp
        or just "user_somedirµtemp" for the config/shortcut file"""
        userScriptDir = api._getConfigDir() + "scripts/"

        if os.path.isdir(userScriptDir):
            #menu = self.ui.menuAdvanced.addMenu('User Commands')
            #menu = self.ui.menubar.addMenu('User Commands')
            menu = self.ui.menuUser_Commands
            subs = {"":menu,} #all submenus. Starts only with the scripts/ menu which has no extra name.

            for rootOriginal, dirs, files in os.walk(userScriptDir):
                root = rootOriginal.replace(api._getConfigDir() + "scripts/", '')
                current = subs[root.split("/")[-1]]

                if dirs:
                    for submenu in dirs:
                        newSubMenu = current.addMenu(submenu)
                        subs[submenu] = newSubMenu

                for name in files:
                    if name.endswith((".py")):
                        pathOriginal = rootOriginal + "/" + name

                        path = pathOriginal.replace(".py", "")
                        path = path.replace("//", "/")
                        path = path.replace(api._getConfigDir() + "scripts/", '')
                        actionName = "user_" + path
                        actionName = actionName.replace("/", "µ") #replace with a seldom symbol. / is not allowed in variable names.
                        entry = current.addAction(actionName)  #just to avoid collisions
                        entry.setText(name.replace(".py", "")) #but use a pretty name.
                        self.connectMenuAction(entry, lambda pathOriginal=pathOriginal: exec(open(pathOriginal).read()))
                        #self.connect(entry,QtCore.SIGNAL('triggered()'), lambda pathOriginal=pathOriginal: exec(open(pathOriginal).read()))
                        exec("self.ui." + actionName + " = entry") #build the variable name from the strings so we can assign shortcuts to it later with our normal method.

                        with open(pathOriginal, 'r') as f: #peak into the first line of the file
                            first_line = f.readline()

                        #Only if this is a normal script file add it to the allActions list which is used to deactivate actions on program start.
                        #If you add #START as first line you can use your script anyway.
                        if not first_line.strip().startswith("#START"):
                            self.allActions.append(eval("self.ui." + actionName))
                        #vars()["self.ui." + actionName] = entry    #somehow this does not work in this scope.
        else:
            warnings.warn("Script dir not found. Creating empty " + userScriptDir + " with README file.")
            os.makedirs(userScriptDir)
            #Create small readme
            f2 = open(api._getConfigDir() + 'scripts/README', "w")
            f2.write("Place script files with .py extensions here. You can use nested directories.\n\nThe script files can use the 'api.foobar()' namespace and the complete python syntax.\nThey work with the visible user cursor. This means what a user can do, a script can do as well.\nConsult the Laborejo manual for more information or read the api.py file for commands.\n\nExample:\nfor i in range(8):\n    api.insert4()\n    api.up()\n")
            f2.close()

    def generateSubstitutionsMenu(self):
        """Generate the program wide substitutions.
        No filespecific data.
        Runs only once add the beginning of the program."""
        self.substitutionsMenu = QtGui.QMenu()
        #tmp = self.ui.menuAdvanced.addMenu(self.substitutionsMenu, )
        tmp = self.ui.menuAdvanced.insertMenu(self.ui.actionRemove_Substitution, self.substitutionsMenu )

        tmp.setText('Laborejo Substitutions')

        alreadyProcessedGroups = {} #each group shall only create one submenu. No nested menus.

        for key, value in sorted(api._getSession().substitutions.items()):  #key is the name string. value is a tuple. (optional group, core.container). We don't care about the container here.
            group = value.group
            if group and not group in alreadyProcessedGroups:
                newSubMenu = self.substitutionsMenu.addMenu(group)
                alreadyProcessedGroups[group] = newSubMenu

            key2 = key.replace("subst_", "")
            pretty = " ".join(key2.split("_")).title()
            if group:
                entry = alreadyProcessedGroups[group].addAction(key)
            else:
                entry = self.substitutionsMenu.addAction(key) #use the real name.
            self.allActions.append(entry)
            entry.setText(pretty) #but display a pretty name.
            #self.connect(entry,QtCore.SIGNAL('triggered()'), lambda key=key: api.substitution(key))
            self.connectMenuAction(entry, lambda key=key: api.substitution(key))

    def closeTabMenu(self):
        """Close the tab through the menu or by a shortcut.
        But not by clicking with the mouse on an [x]"""
        idx = self.ui.Tab.currentIndex()
        self.closeTabWarning(idx)

    def closeTabWarning(self, tabIndex):
        """Check if there is unsaved data in a movement before closing
        the tab."""
        t = self.ui.Tab
        t.setCurrentIndex(tabIndex)
        if not tabIndex == t.currentIndex():
            raise IndexError("Tab to close is not current tab eventhough it was changed beforehand")
        if self.session.isCurrentClean():
            del self.session.scenes[self.ui.Tab.currentWidget().data]
            api.close()
            t.removeTab(tabIndex)
        else:
            tabname = t.tabText(tabIndex)
            reply = QtGui.QMessageBox.question(self, 'Unsaved Data', "Unsaved Data in " + tabname + ".\nHow to proceed?", QtGui.QMessageBox.Save | QtGui.QMessageBox.Discard | QtGui.QMessageBox.Cancel , QtGui.QMessageBox.Cancel)
            if reply == QtGui.QMessageBox.Discard:
                del self.session.scenes[self.ui.Tab.currentWidget().data]
                self.session.setSavestatusClean()
                api.close()
                t.removeTab(tabIndex) #Just close
            elif reply == QtGui.QMessageBox.Save:
                if self.save(): #This calls the gui dialog if there is no filename. Could be canceled as well.
                    del self.session.scenes[self.ui.Tab.currentWidget().data]
                    self.session.setSavestatusClean()
                    api.close()
                    t.removeTab(tabIndex)
                else: #If canceled just safely return to the unsaved data.
                    pass
            else: #Cancel
                pass
        if not self.session.scenes: #it is empty? No files left open?
            self.switchActionsOff()
            self.placeholderMovementSubstitutions.clear()
            self.placeholderMovementContainers.clear()

    def closeEvent(self, event): #does not need to be assigned or connected. This overwrites a default qt method.
        """Check if there is any unsaved data in the session before
        closing the program"""
        if self.session.saveStatus: #There is dirty data.
            #Offer to cancel, Quit Anyway or SaveAll and then Quit.
            reply = QtGui.QMessageBox.question(self, 'Unsaved Data', "One or more files are unsaved. How to proceed?\nWarning: Unnamed Files will be discarded.", QtGui.QMessageBox.SaveAll | QtGui.QMessageBox.Discard | QtGui.QMessageBox.Cancel , QtGui.QMessageBox.Cancel)

            if reply == QtGui.QMessageBox.Discard:
                event.accept() #Just quit
            elif reply == QtGui.QMessageBox.SaveAll:
                api.saveAll() #this never fails. It has no user interaction. No test necessarry.
                event.accept()
            else: #Cancel
                event.ignore()
        else: #Everything is clean and saved. Good Bye.
            event.accept()

    #Save, Load, Export
    #####
    def new(self, backendWorkspace):
        """create a new tab with an empty, unsaved score"""
        #backendWorkspace = api.new()
        if backendWorkspace.lastKnownFilename:
            fileName = backendWorkspace.lastKnownFilename
        else:
            fileName = "unnamed"
        workspace = GuiMovement(backendWorkspace)
        self.session.scenes[backendWorkspace] = workspace #This must be done early. loadBackendData already depends on existing session.foo vars.
        workspace.loadBackendData() #This is not load a file but create the gui from the backend data.
        workspace.fileName = ""
        tabIndex = self.ui.Tab.addTab(workspace, fileName)
        self.ui.Tab.setCurrentIndex(tabIndex)
        c = self.randomColor()
        pixmap = QtGui.QPixmap(20, 20)
        pixmap.fill(QtGui.QColor(c))
        workspace.icon = QtGui.QIcon(pixmap)
        self.ui.Tab.tabBar().setTabIcon(tabIndex, workspace.icon)
        workspace.updateGui() #Update Track Properties etc.
        workspace.cursorPitch.update()
        workspace.cursorPosition.update()
        self.session.setSavestatusClean()
        self.switchActionsOn() #We have at least one file now, everything is allowed.

    def load(self, directLoad = None, asTemplate = False):
        """Present the gui load dialog,
        ask the api to open the filepath"""

        if directLoad:
            fileNameList = [directLoad]
        else:
            fileNameList = QtGui.QFileDialog.getOpenFileNames(self, "Open File(s)", self.last_open_dir, "Laborejo Files (*.lbj *.lbjs);;All Files (*.*)")

        if fileNameList: #not cancelled
            atLeastOneFileOpened = False

            #Special Case: There is only one file open, it is unnamed but savestatus is clean. This means it is a new, unmodified file.
            #In that case close that empty file.
            if len(self.session.scenes.items()) == 1 and (not self.session.saveStatus) and self.session.currentGuiWorkspace.fileName is "":
                self.closeTabWarning(0) #despite the name this will not result in any GUI or warning. SaveStatus is all good.

            for fileName in fileNameList:
                #first check if this file is already open in this GUI. Templates can always be opened.
                if self.session.isFilenameAvailable(fileName) or asTemplate:
                    backendWorkspace = api.load(fileName, force = False, asTemplate = asTemplate) #this returns False if the user chose a non-laborejo file. Only use for testing or emergency.
                    if backendWorkspace:
                        workspace = GuiMovement(backendWorkspace) #Before creating a new gui workspace we have to delete the old Graphics, including the cursors.
                        self.session.scenes[backendWorkspace] = workspace #This must be done early. loadBackendData already depends on existing session.foo vars.
                        workspace.loadBackendData()
                        if asTemplate:
                            # The backend does not know about templates. It just has open and new. New has no filename, open has one. A template is a gui construct which gives a "new" file to the user which already has content.
                            workspace.fileName = ""
                            backendWorkspace.lastKnownFilename = "" #We switch the backend name to "new" here.
                            tabIndex = self.ui.Tab.addTab(workspace, "unnamed")
                        else:
                            workspace.fileName = fileName
                            tabIndex = self.ui.Tab.addTab(workspace, fileName)
                            self.addRecentFiles(fileName)
                        self.ui.Tab.setCurrentIndex(tabIndex)
                        c = self.randomColor()
                        pixmap = QtGui.QPixmap(20, 20)
                        pixmap.fill(QtGui.QColor(c))
                        workspace.icon = QtGui.QIcon(pixmap)
                        self.ui.Tab.tabBar().setTabIcon(tabIndex, workspace.icon)
                        workspace.updateGui()
                        workspace.cursorPitch.update()
                        workspace.cursorPosition.update()
                        self.session.setSavestatusClean()
                        atLeastOneFileOpened = True
                        self.last_open_dir = os.path.dirname(fileName)
                else:
                    warnings.warn("File " +  fileName +  "is already open")

            if atLeastOneFileOpened:
                self.switchActionsOn() #We have at least one file now, everything is allowed.

    def loadAsTemplate(self, directLoad = None):
        self.load(directLoad, asTemplate = True)


    def previewPDF(self):
        api.previewPDF(pdfviewer = config.pdfviewer)


    def substitutionsMovementFromFile(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self, "Choose File to convert to Substitutions", self.last_subst_dir, "Laborejo Files (*.lbj *.lbjs);;All Files (*.*)")
        if fileName:
            api.substitutionsMovementFromFile(fileName)
            self.last_subst_dir = os.path.dirname(fileName)

    def substitutionsMovementToFile(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self, "Save Laborejo Substitutions as Movement.", self.last_save_dir , "Laborejo Movement (*.lbjs)")
        if fileName:
            self.last_subst_dir = os.path.dirname(fileName)
            api.substitutionsMovementToFile(filepath = fileName) #This opens a new file/tab

    def save(self, saveAs = False):
        """The gui save dialog to get a filepath or direct save.
        Save is always for the current active workspace/movement"""
        currentWs = self.session.currentGuiWorkspace
        if saveAs or not currentWs.fileName:
            fileName = QtGui.QFileDialog.getSaveFileName(self, "Save Laborejo Movement file.", self.last_save_dir , "Laborejo Movement (*.lbjs)")
            if fileName: #Not pressed cancel?
                if not os.path.splitext(fileName)[1]: #no file extension given?
                    fileName = fileName + ".lbjs"
                api.save(fileName)
                currentWs.fileName = fileName
                self.ui.Tab.setTabText(self.ui.Tab.currentIndex(), fileName)
                self.addRecentFiles(fileName)
                self.session.setSavestatusClean() #The currently active file will be set clean.
                self.last_save_dir = os.path.dirname(fileName)
                return True
            else: #User pressed cancel
                return False
        elif currentWs.fileName:
            api.save(currentWs.fileName)
            self.session.setSavestatusClean() #The currently active file will be set clean.
            return True

    def saveAs(self):
        self.save(saveAs = True)

    def exportOptions(self):
        """Edit export options for the currently activated file/tab."""
        #a = os.listdir(os.path.dirname(__file__) + "/../backend/lilypondtemplates/")

        a = list(api.getTemplateList())

        if api._getWorkspace().lytemplate in a:
            a.insert(0, api._getWorkspace().lytemplate) #use the current template as default position for the template gui.
            warning = ""
        else:
            a.insert(0, a[0]) #duplicate the first item to make this a fedit compatible list.
            warning = "Warning: Template " + api._getWorkspace().lytemplate + " could not be found!\nChoose a different one now or press cancel.\n\n"
        dictionary = {"template": ("Lilypond/PDF Template", a), "fontsize": ("Lilypond font size", api._getMovement().lyFontSize)}
        ret = fedit(dictionary, "Export Options", warning + "Modifiy export behaviour for PDF, Midi and Lilypond files.\n\nThese options are saved in your Laborejo save file.")
        if ret: #not pressed cancel?
            api._getWorkspace().lytemplate = ret["template"]
            api._getMovement().lyFontSize = ret["fontsize"]

    def exportMidi(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self, "Export Movement as midi. Give full name including file extension.", self.last_export_dir, "*.mid *.midi *.MID *.MIDI")
        if fileName:
            api.exportMidi(fileName)
            self.last_export_dir = os.path.dirname(fileName)

    def exportMidiTracks(self):
        dirName = QtGui.QFileDialog.getExistingDirectory(self, caption = "Export each track as midi file. Choose a directory.", directory = self.last_export_dir)
        if dirName:
            api.exportMidiAsTracks(dirName)
            self.last_export_dir = dirName

    def exportArdour(self):
        dirName = QtGui.QFileDialog.getExistingDirectory(self, caption = "Export Ardour 3 Session. Choose a directory where ANOTHER directory will be created.", directory = self.last_export_dir)
        if dirName:
            api.exportArdourThree(dirName)
            self.last_export_dir = dirName

    def exportLilypond(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self, "Export complete Laborejo Session as Lilypond. Give full name including file extension.", self.last_export_dir, "*.ly")
        if fileName:
            api.exportLilypond(fileName)
            self.last_export_dir = os.path.dirname(fileName)

    def exportPDF(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self, "Export complete Laborejo Session as PDF (Lilypond)", self.last_export_dir, "*.pdf")
        if fileName:
            api.exportPDF(fileName)
            self.last_export_dir = os.path.dirname(fileName)

    def viewLilypondText(self):
        if self.ui.lilyfloat.isVisible():
            self.ui.lilyfloat.hide()
        else:
            self.ui.lilyfloat.setText(api.generateLilypond())
            self.ui.lilyfloat.show()

    def importLisalo(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self, "Import Lisalo File", self.last_import_dir, "Lisalo Files (*.lsl);;All Files (*.*)")
        if fileName:
            api.importLisalo(fileName)
            self.last_import_dir = os.path.dirname(fileName)

    def importLisaloBus(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self, "Import Lisalo File", self.last_import_dir, "Lisalo Files (*.lsl);;All Files (*.*)")
        if fileName:
            api.importLisalo(fileName, trackPerBus = True)
            self.last_import_dir = os.path.dirname(fileName)


    def randomColor(self):
        """Generates a nice pastell color"""
        def newColor():
            red, green, blue = randint(1, 255), randint(1, 255), randint(1, 255)
            #red, green, blue = 205 + random() * 50, 205 + random() * 50, 205 + random() * 50,
            alpha = 91 + random() * 100
            return (red, green, blue, alpha)
        c = newColor()
        while c in self.colors:
            c = new()
        self.colors.append(c)
        return QtGui.QColor(c[0], c[1], c[2], 255)

class MusicTab(QtGui.QWidget):
    """The ui part of a GuiMovement"""

    def __init__(self):
        super(MusicTab, self).__init__()
        self.setParent(main.ui.Tab)
        self.ui = Ui_musicTab()  #self.ui is the node to all qt widgets in this tab.
        self.ui.setupUi(self)
        self.rubberBandOrigin = None #the temporary rubberband origin
        self.rubberBand = None #the temporary rubberband
        self.icon = None # Groups and Movements use the same icon/color.

        workingAreaWidget = self.ui.score #The QGraphicsScene
        self.orgFocusIn = workingAreaWidget.focusInEvent
        self.orgFocusOut = workingAreaWidget.focusOutEvent
        workingAreaWidget.focusInEvent = self.customFocusIn
        workingAreaWidget.focusOutEvent = self.customFocusOut
        workingAreaWidget.setFocusPolicy(QtCore.Qt.WheelFocus)

        #TODO: self.ui.score.setFocus(True) Does not work in the beginning. We need to click in the score with the mouse at least once!

        #Setup the main notation view, the score, by adding a Scene.
        self.graphicsViewRect = [0, 0, 0 , 0] #x, y, width, height
        self.scene = QtGui.QGraphicsScene()
        self.scene.setBackgroundBrush(QtGui.QColor("white"))
        self.ui.score.setScene(self.scene)
        self.updateSceneRect()

        self.ui.score.mousePressEvent = self.mousePressEventCustom
        self.ui.score.mouseMoveEvent = self.mouseMoveEventCustom
        self.scene.mouseReleaseEvent = self.mouseReleaseEventCustom
        self.scene.mouseDoubleClickEvent = self.mouseReleaseEventCustom
        self.wheelEventOrg = self.ui.score.wheelEvent
        self.ui.score.wheelEvent = self.mouseWheelEventCustom
        self.scene.contextMenuEvent = self.contextMenuEventCustom

        #Setup the synced incipit frame left of the score for group and other meta data.
        self.incipitScene = QtGui.QGraphicsScene()
        self.incipitScene.setBackgroundBrush(QtGui.QColor("white"))
        self.ui.incipit.setScene(self.incipitScene)

        #Sync the vertical incipit scrollbar (which is never shown) with the score scrollbar.
        self.incipitVerticalBar = self.ui.incipit.verticalScrollBar()
        #self.scoreVerticalBar =  self.ui.score.verticalScrollBar()
        self.ui.score.setVerticalScrollBar(self.incipitVerticalBar)

        #Initial Zoom Levels
        newTransform = QtGui.QTransform()
        newTransform.scale(0.7, 0.7)
        self.ui.score.setTransform(newTransform)
        newTransform2 = QtGui.QTransform()
        newTransform2.scale(1, 0.7)
        self.ui.incipit.setTransform(newTransform2)

        #Minimap
        self.ui.minimap.hide() #hidden by default
        self.ui.minimap.setScene(self.scene)
        self.scene.sceneRectChanged.connect(lambda newRect: self.ui.minimap.fitInView(newRect, QtCore.Qt.KeepAspectRatio))
        self.ui.minimap.setViewportUpdateMode(QtGui.QGraphicsView.FullViewportUpdate)

        #Deactivate minimap mouse buttons which were originally for the whole scene.
        def nothingMouse(ev):
            pass
        self.ui.minimap.mouseReleaseEvent = lambda ev: self.ui.score.centerOn(self.ui.minimap.mapToScene(ev.pos()))
        self.ui.minimap.mousePressEvent = nothingMouse
        self.ui.minimap.contextMenuEvent = nothingMouse
        self.ui.minimap.mouseDoubleClickEvent = nothingMouse

        #we update the viewport manually
        #but some actions trigger an update, like scrollbars.
        workingAreaWidget.setViewportUpdateMode(QtGui.QGraphicsView.NoViewportUpdate)
        self.ui.score.horizontalScrollBar().valueChanged.connect(self.ui.score.viewport().update)
        self.ui.score.verticalScrollBar().valueChanged.connect(self.ui.score.viewport().update) #slider move is the manual slider movement.


        #Track Properties
        self.ui.trackUniqueName.textEdited.connect(lambda: api.trackName(self.ui.trackUniqueName.text())) #The api function will send back the change to the trackProperties listener which updates the gui.
        self.ui.trackGroup.textEdited.connect(lambda: self.changeGroup(self.ui.trackGroup.text()))
        self.ui.trackLongInstrumentName.textEdited.connect(lambda: api.trackChangeAttribute("longInstrumentName", self.ui.trackLongInstrumentName.text()))
        self.ui.trackShortInstrumentName.textEdited.connect(lambda: self.changeNameMarker(self.ui.trackShortInstrumentName.text()))
        self.ui.trackVoice.stateChanged.connect(lambda: api.trackChangeAttribute("mergeWithUpper", self.ui.trackVoice.checkState()))
        self.ui.trackDirectives.clicked.connect(lambda: api.directivesGui(api._getTrack()))
        self.ui.trackAmbitus.stateChanged.connect(lambda: api.trackChangeAttribute("ambitus", self.ui.trackAmbitus.checkState()))
        self.ui.trackLyricsAbove.stateChanged.connect(lambda: api.trackChangeAttribute("forceLyricsAbove", self.ui.trackLyricsAbove.checkState()))
        self.ui.trackHideLilypond.stateChanged.connect(lambda: api.trackChangeAttribute("hideLilypond", self.ui.trackHideLilypond.checkState()))
        self.ui.trackNrOfLines.valueChanged.connect(lambda: api.trackChangeAttribute("nrOfLines", self.ui.trackNrOfLines.value()))

        self.ui.trackTranposeFrom.addItems(api.lilypond.sortedNoteNameList)
        self.ui.trackTranposeTo.addItems(api.lilypond.sortedNoteNameList)
        self.ui.trackTranposeFrom.view().setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum) #make sure the list is visible, always.
        self.ui.trackTranposeTo.view().setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)#make sure the list is visible, always.
        self.ui.trackTranposeFrom.setCurrentIndex(api.lilypond.sortedNoteNameList.index(api.lilypond.pitch2ly[api._getTrack().transposition[0]]))
        self.ui.trackTranposeTo.setCurrentIndex(api.lilypond.sortedNoteNameList.index(api.lilypond.pitch2ly[api._getTrack().transposition[1]]))
        self.ui.trackTranposeFrom.currentIndexChanged.connect(lambda: api.trackChangeAttribute("transposition", (api.lilypond.ly2pitch[api.lilypond.sortedNoteNameList[self.ui.trackTranposeFrom.currentIndex()]], api.lilypond.ly2pitch[api.lilypond.sortedNoteNameList[self.ui.trackTranposeTo.currentIndex()]]))) # :)
        self.ui.trackTranposeTo.currentIndexChanged.connect(lambda: api.trackChangeAttribute("transposition", (api.lilypond.ly2pitch[api.lilypond.sortedNoteNameList[self.ui.trackTranposeFrom.currentIndex()]], api.lilypond.ly2pitch[api.lilypond.sortedNoteNameList[self.ui.trackTranposeTo.currentIndex()]]))) # ;)

        self.ui.trackSize.addItems(["normal", "big", "small", "tiny"]) #Sizes List
        self.ui.trackSize.view().setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self.ui.trackSize.currentIndexChanged.connect(lambda: api.trackChangeAttribute("lilypondSize", self.ui.trackSize.currentText()))

        self.ui.trackVoicePreset.addItems(dictionaries.voicePresets) #Voice Presets List
        self.ui.trackVoicePreset.currentIndexChanged.connect(lambda: api.trackChangeAttribute("voicePreset", self.ui.trackVoicePreset.currentText()))

        self.ui.trackMute.stateChanged.connect(lambda: api.trackChangeAttribute("smfMute", self.ui.trackMute.checkState()))
        self.ui.trackSolo.stateChanged.connect(lambda: api.trackChangeAttribute("smfSolo", self.ui.trackSolo.checkState()))

        self.ui.trackChannelMin.valueChanged.connect(lambda: api.trackChangeAttribute("smfChannel", [self.ui.trackChannelMin.value() -1, self.ui.trackChannelMin.value() -1])) #In the GUI we keep both values seperate, but the backend safes both as list [min, max]
        self.ui.trackMidiTransposition.valueChanged.connect(lambda: api.trackChangeAttribute("smfTransposition", self.ui.trackMidiTransposition.value()))
        self.ui.trackVolume.valueChanged.connect(lambda: api.trackChangeAttribute("smfVolume", self.ui.trackVolume.value() -1))
        self.ui.trackPan.valueChanged.connect(lambda: api.trackChangeAttribute("smfPan", self.ui.trackPan.value() -1))
        self.ui.trackInstrument.addItems(api.playback.instrumentList) #Instrument List
        self.ui.trackInstrument.view().setSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self.ui.trackInstrument.currentIndexChanged.connect(lambda: api.trackChangeAttribute("smfPatch", self.ui.trackInstrument.currentIndex()))
        self.ui.trackBank.valueChanged.connect(lambda: api.trackChangeAttribute("smfBank", self.ui.trackBank.value() -1))

        self.ui.trackJackmidiPort.textEdited.connect(lambda: api.trackChangeAttribute("jackPort", self.ui.trackJackmidiPort.text()))
        self.ui.jackChannelMin.valueChanged.connect(lambda: api.trackChangeAttribute("jackChannel", [self.ui.jackChannelMin.value() -1, self.ui.jackChannelMax.value() -1])) #A change of one frontend channel parameter changes both backend min, max values.
        self.ui.jackChannelMax.valueChanged.connect(lambda: api.trackChangeAttribute("jackChannel", [self.ui.jackChannelMin.value() -1, self.ui.jackChannelMax.value() -1])) #A change of one frontend channel parameter changes both backend min, max values.
        self.ui.jackPatch.valueChanged.connect(lambda: api.trackChangeAttribute("jackPatch", self.ui.jackPatch.value() -1))
        self.ui.jackBank.valueChanged.connect(lambda: api.trackChangeAttribute("jackBank", self.ui.jackBank.value() -1))
        self.ui.jackVolume.valueChanged.connect(lambda: api.trackChangeAttribute("jackVolume", self.ui.jackVolume.value() -1))
        self.ui.jackPan.valueChanged.connect(lambda: api.trackChangeAttribute("jackPan", self.ui.jackPan.value() -1))
        self.ui.jackTranspose.valueChanged.connect(lambda: api.trackChangeAttribute("jackTransposition", self.ui.jackTranspose.value()))
        self.ui.jackAudioL.textEdited.connect(lambda: api.trackChangeAttribute("jackAudioReturnPortLeft", self.ui.jackAudioL.text()))
        self.ui.jackAudioR.textEdited.connect(lambda: api.trackChangeAttribute("jackAudioReturnPortRight", self.ui.jackAudioR.text()))

        #Lyric Frame / TextEdit
        self.ui.lyrics.hide() #Lyrics are hidden by default
        def changeLyrics(): #What happens when the lyric frame is changed.
            api._getTrack().lyrics = self.ui.lyricTextEdit.toPlainText()
            self.currentTrack.drawLyrics() #eventhough we are in MusicTab we are also in GuiMovement since it is a child class. The only child class.
            self.ui.score.viewport().update() #Update the grahipcsview manually

        self.ui.lyricTextEdit.textChanged.connect(changeLyrics)
        self.lyricTextEditOriginalKeyPressEvent = self.ui.lyricTextEdit.keyPressEvent
        self.ui.lyricTextEdit.keyPressEvent = self.lyricKeyPressInterceptor

        #Substitutions
        self.movementSubstitutionsMenu = None #an everchanging menu of substitutions
        self.movementContainerMenu = None #an everchanging menu of container

    def customFocusIn(self, event):
        for action in main.genericKeyActions:
            action.setEnabled(True)
        self.orgFocusIn(event)

    def customFocusOut(self, event):
        for action in main.genericKeyActions:
            action.setDisabled(True)
        self.orgFocusOut(event)

    def lyricKeyPressInterceptor(self, qKeyEvent):
        if qKeyEvent.nativeModifiers() and qKeyEvent.nativeVirtualKey() == 32: #shift and space
            self.ui.lyricTextEdit.insertPlainText(" -- ")
        self.lyricTextEditOriginalKeyPressEvent(qKeyEvent)

    def updateGui(self):
        """Update several frame and widgets that derive their appearance
        from backend data"""
        self.updateTrackProperties()
        self.updateLyricsFrame()

    def updateStatusBar(self):
        cursor = api._getActiveCursor()
        msg = ("Pitch: " + api.lilypond.pitch2ly[api.pitch.toScale(cursor.pitchindex, cursor.prevailingKeySignature[-1])]
            + " || " + "Track: " + str(api._getTrackIndex() +1 )
            + " Bar: " + str(cursor.measureNumber + 1)
            +  " Pos: " + str(round(cursor.metricalPosition +1, 4))
            + ". Sec: " + str(round(cursor.timeindex, 4))
            + " TickIndex: " + str(cursor.tickindex)
            + " || Prevailing Duration: "
            + api.lilypond.dur2ly[cursor.prevailingDuration]
            + " || Mode: " + main.mode
            + " Type: " + cursor.get_item().__class__.__name__
            + " Stack: " + str(api._getIndex()))
        main.statusBar().showMessage(msg + " || " + main.more)

    def updateTrackProperties(self):
        """Uses the (backend) cursor position.
        Each of the following updates will trigger a change signal,
        defined in MusicTab.__init__(), which will in return trigger
        a re-set of the backend value.
        In most cases this is just a waste of resources, in cases like
        JackMidiChannel where a single backend value influences two
        frontend values this leads to value errors.

        We can't prevent the Qt Signal flow but we can prevent the value
        changing in the backend by temporarily deactivating the api
        function.
        """
        org = api.trackChangeAttribute
        def empty(*args):
            pass
        api.trackChangeAttribute = empty  #At the end of this function the original function gets restored.

        track = api._getTrack() #backend instance
        if track.group:
            nameLabel =str(api._getTrackIndex() +1 ) + ": " +  track.uniqueContainerName + "(" + track.group + ")"
        else:
            nameLabel = str(api._getTrackIndex() +1 ) + ": " +  track.uniqueContainerName
        self.ui.trackPropertiesLabel.setText(nameLabel)
        self.ui.trackNrOfLines.setValue(track.nrOfLines)
        self.ui.trackUniqueName.setText(track.uniqueContainerName)
        self.ui.trackGroup.setText(track.group)
        self.ui.trackShortInstrumentName.setText(track.shortInstrumentName)
        self.ui.trackLongInstrumentName.setText(track.longInstrumentName)
        self.ui.trackTranposeFrom.setCurrentIndex(api.lilypond.sortedNoteNameList.index(api.lilypond.pitch2ly[track.transposition[0]]))
        self.ui.trackTranposeTo.setCurrentIndex(api.lilypond.sortedNoteNameList.index(api.lilypond.pitch2ly[track.transposition[1]]))
        self.ui.trackHideLilypond.setChecked(track.hideLilypond)
        self.ui.trackLyricsAbove.setChecked(track.forceLyricsAbove)
        self.ui.trackVoice.setChecked(track.mergeWithUpper)
        self.ui.trackMute.setChecked(track.smfMute)
        self.ui.trackSolo.setChecked(track.smfSolo)
        self.ui.trackSize.setCurrentIndex(["normal", "big", "small", "tiny"].index(track.lilypondSize))
        self.ui.trackVoicePreset.setCurrentIndex(dictionaries.voicePresets.index(track.voicePreset))
        directives = len(track.directivePre) + len(track.directiveMid) + len(track.directivePst) + len(track.instructionPre) + len(track.instructionPst) #total number of directives
        self.ui.trackDirectives.setText(str(directives) +  " Directives")
        self.ui.trackAmbitus.setChecked(track.ambitus)

        self.ui.trackVolume.setValue(track.smfVolume + 1)
        self.ui.trackPan.setValue(track.smfPan +1)
        self.ui.trackInstrument.setCurrentIndex(track.smfPatch)
        self.ui.trackChannelMin.setValue(track.smfChannel[0] + 1)
        self.ui.trackMidiTransposition.setValue(track.smfTransposition)
        self.ui.trackBank.setValue(track.smfBank + 1)

        self.ui.trackJackmidiPort.setText(track.jackPort)
        self.ui.jackChannelMin.setValue(track.jackChannel[0] + 1)
        self.ui.jackChannelMax.setValue(track.jackChannel[1] + 1)
        self.ui.jackPatch.setValue(track.jackPatch + 1)
        self.ui.jackVolume.setValue(track.jackVolume + 1)
        self.ui.jackPan.setValue(track.jackPan + 1)
        self.ui.jackTranspose.setValue(track.jackTransposition)
        self.ui.jackBank.setValue(track.jackBank + 1)
        self.ui.jackAudioL.setText(track.jackAudioReturnPortLeft)
        self.ui.jackAudioR.setText(track.jackAudioReturnPortRight)

        api.trackChangeAttribute = org

    def updateLyricsFrame(self):
        """Called at each track change"""
        track = api._getTrack() #backend instance
        self.ui.lyrics.setTitle("Lyrics - Track " + track.uniqueContainerName)
        self.ui.lyricTextEdit.setPlainText(track.lyrics)

    def updateSceneRect(self):
        """x, y, width, height"""
        self.ui.score.setSceneRect(self.graphicsViewRect[0], -60, self.graphicsViewRect[2] + 100, self.graphicsViewRect[3] + 100)
        self.ui.incipit.setSceneRect(self.graphicsViewRect[0], -59, 50, self.graphicsViewRect[3] + 100) #50 because the view is fixed to 50 pixels
        self.ui.score.viewport().update()


    def visibleRect(self):
        topLeft = QtCore.QPointF(self.ui.score.mapToScene( 0, 0 ))
        bottomRight = QtCore.QPointF(self.ui.score.mapToScene(self.ui.score.viewport().width() - 1, self.ui.score.viewport().height() - 1 ))
        rect = QtCore.QRectF(topLeft, bottomRight)
        return rect

    def findGuiItemsFromPosCrawlLeft(self, parameterPos):
        #init
        clickPos = parameterPos
        x = clickPos.x()
        y = clickPos.y()
        directHit = False
        #First find the track.
        #tupl = (lowest pos, highest pos, gui-track) #'lowest' means highest in gui coordinates. 0 is the top scene border
        guiTrack = [tupl for tupl in self.cachedTrackPositions if tupl[0] < y < tupl[1]][0][2]

        #Create a rectangle/beam that crawls left and returns all items within.
        #We take the last one
        qRectangleF = QtCore.QRectF(x-200, y-20, 200, 45)   #shift-x, shift-y, width, height  . If you shift the middle of the max height/width you have the cursor in the middle.

        #global draw
        #try: self.scene.removeItem(draw)
        #except: pass
        #draw = QtGui.QGraphicsRectItem(qRectangleF)
        #self.scene.addItem(draw)

        foundItems = self.scene.items(qRectangleF)
        foundItemsFilter = list(filter(lambda item: isinstance(item, guiitems.GuiItem) and item.track is guiTrack, foundItems))

        if list(filter(lambda item: isinstance(item, guiitems.GuiItem), self.scene.items(x-3, y-10, 6, 30))):
            directHit = True

        if foundItemsFilter: #Item left of the click found.
            foundItem  = sorted(foundItemsFilter, key=lambda x: x.cachedIndex)[-1] #the highest cached flat cursor index is the topmost right item.
            foundItemIndex = foundItem.cachedIndex
        elif len(guiTrack.data.container) == 1: #empty track, only appending. Hopefully the backend never changes "Only Appending = Empty".
            foundItem = PSEUDO #Appending
            foundItemIndex = 0
        else:
            return (None, None, None, None)

        return (foundItem, foundItemIndex, guiTrack.data, directHit)

    def mouseReleaseEventCustom(self,ev):
        """Create a virtual "big" cursor that is not a point
        but a rectangle.
        Step by step increase the search rectangle until you find
        something or the left border, which is 0."""
        modifiers = QtGui.QApplication.keyboardModifiers()

        if ev.button() == QtCore.Qt.LeftButton:
            self.rubberBand.hide()

            #Rubberband selection
            #does not move the cursor.
            if self.rubberBand.size().height() * self.rubberBand.size().width() > 100: #treshhold to not trigger a rubberband selection by accident
                viewCoordinatesRect = self.rubberBand.geometry()
                if modifiers == QtCore.Qt.ControlModifier:
                    self.deselectInRectangle((self.ui.score.mapToScene(viewCoordinatesRect)).boundingRect())
                else:
                    self.selectInRectangle((self.ui.score.mapToScene(viewCoordinatesRect)).boundingRect())

            #Single Click
            #Not a rubberband.
            else:
                foundItem, foundItemIndex, backendTrack, directHit = self.findGuiItemsFromPosCrawlLeft(parameterPos = ev.scenePos())
                if backendTrack:
                    #Ctrl+Click. Add one Item to the selection
                    if modifiers == QtCore.Qt.ControlModifier:
                        try: #items like the MultiMeasureRest and Upbeat hide their data for the caching engine to work.
                            api.remoteToggleSelect(foundItem.dataOrg, backendTrack)
                        except:
                            api.remoteToggleSelect(foundItem.data, backendTrack)


                    #Shift+Click. Add a rectangle of items to the selection
                    elif modifiers == QtCore.Qt.ShiftModifier:
                        #We create a rectangle from the old cursor position and the new. From then:
                        #Is the same as rectangle selection plus moving the cursor to the target position afterwards.
                        #this way we get a multitrack selection. Something that the api alone cannot do.
                        originCursorPosX = self.cursorPosition.pos().x()
                        originCursorPosY = self.cursorPosition.pos().y()

                        api.cursorFlatPosition((api._getIndexByTrack(backendTrack), foundItemIndex))
                        if not directHit:
                            api.right() #we search left but want in the space 'between' the notes. This is already protected against going right at the end of the track.

                        targetCursorPosX = self.cursorPosition.pos().x()
                        targetCursorPosY = self.cursorPosition.pos().y()

                        if originCursorPosX <= targetCursorPosX: #the rectangle goes from left to right
                            if originCursorPosY <= targetCursorPosY: #from top to bottom
                                self.selectInRectangle(QtCore.QRectF(originCursorPosX+20, originCursorPosY, targetCursorPosX-originCursorPosX-20, originCursorPosY + targetCursorPosY + 80)) #build a rectangle from the current and old position and let the function create a selection from it.
                            else: #bottom to top
                                self.selectInRectangle(QtCore.QRectF(originCursorPosX+20, targetCursorPosY, targetCursorPosX-originCursorPosX-20, originCursorPosY + targetCursorPosY + 80)) #build a rectangle from the current and old position and let the function create a selection from it.
                        else: #backwards
                            if originCursorPosY <= targetCursorPosY: #from top to bottom
                                self.selectInRectangle(QtCore.QRectF(targetCursorPosX+20, targetCursorPosY, originCursorPosX-targetCursorPosX-20, 80))
                            else: #bottom to top
                                self.selectInRectangle(QtCore.QRectF(targetCursorPosX+20, targetCursorPosY, originCursorPosX-targetCursorPosX-20, originCursorPosY + targetCursorPosY + 80)) #build a rectangle from the current and old position and let the function create a selection from it.

                    #No modifiers. Move the cursor
                    else:
                        api.cursorFlatPosition((api._getIndexByTrack(backendTrack), foundItemIndex))
                        if not directHit:
                            api.right() #we search left but want in the space 'between' the notes. This is already protected against going right at the end of the track.

                #User clicked outside of any reasonable perimeter.
                #else:
                #    return None #nothing happens.

        #elif ev.button() == QtCore.Qt.RightButton:
            #nothing happens. Context menu is already done on press, not release.
        elif ev.button() == QtCore.Qt.MidButton:
            api.deselectMovement()
        self.ui.score.viewport().update() #Update the grahipcsview manually

    def mouseWheelEventCustom(self, ev):
        modifiers = QtGui.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier | QtCore.Qt.ShiftModifier:
            if ev.delta() > 0:
                self.widen()
            else:
                self.shrinken()
        elif modifiers == QtCore.Qt.ControlModifier:
            if ev.delta() > 0:
                self.zoomIn()
            else:
                self.zoomOut()
        else:
            self.wheelEventOrg(ev)
        self.ui.score.viewport().update()

    def mousePressEventCustom(self, ev):
        """Mouse Press is either the beginning of a 'single click'
        to set the cursor position or to toggleSelect a single item.
        Or it is the beginning of a rubberband """
        if ev.button() == QtCore.Qt.LeftButton:
            self.rubberBandOrigin = ev.pos()
            self.rubberBand = QtGui.QRubberBand(QtGui.QRubberBand.Rectangle, parent = self.ui.score)
            self.rubberBand.setGeometry(QtCore.QRect(self.rubberBandOrigin, QtCore.QSize()))
            self.rubberBand.show()

    def mouseMoveEventCustom(self, ev):
         if self.rubberBand:
            self.rubberBand.setGeometry(QtCore.QRect(self.rubberBandOrigin, ev.pos()).normalized())

    def contextMenuEventCustom(self, ev):
        foundItem, foundItemIndex, backendTrack, directHit = self.findGuiItemsFromPosCrawlLeft(parameterPos = ev.scenePos())
        if backendTrack and not foundItem is PSEUDO:
            api.cursorFlatPosition((api._getIndexByTrack(backendTrack), foundItemIndex))
            if not directHit:
                api.right() #we search left but want in the space 'between' the notes. This is already protected against going right at the end of the track.
        else: #User clicked outside of any reasonable perimeter.
            return None #nothing happens.

        #We are now at the cursor position. Decide which context menu is correct.
        cur = self.currentTrack.getCurrentItem()
        if type(cur) is guiitems.GuiChord:
            tmpMenu = QtGui.QMenu()
            tmpMenu.addMenu(self.movementContainerMenu)
            tmpMenu.addSeparator()
            tmpMenu.addMenu(main.placeholderMovementSubstitutions)
            tmpMenu.addMenu(main.substitutionsMenu)
            tmpMenu.addAction(main.ui.actionCreate_Substitution_from_Selection)
            tmpMenu.addAction(main.ui.actionRemove_Substitution)
            #tmpMenu.exec_(self.ui.score.mapToGlobal(ev.scenePos().toPoint()))  #this needs the cursor position to be correct or a selection. Normal backend behaviour, but prevents us from using a hypothetical substitutionsRemote() which is maybe better that way. The cursor should always move.
            tmpMenu.exec_(ev.screenPos())  #this needs the cursor position to be correct or a selection. Normal backend behaviour, but prevents us from using a hypothetical substitutionsRemote() which is maybe better that way. The cursor should always move.
        else:
            return None

    def selectInRectangle(self, qRectangleF):
        """Used in eventFilter for rubberband selection and
        mouseButtonPressed for Shift+Press"""
        #draw = QtGui.QGraphicsRectItem(qRectangleF)
        #self.scene.addItem(draw)
        rectangledItems = self.scene.items(qRectangleF, mode = QtCore.Qt.IntersectsItemShape)
        filtered = list(filter(lambda item: isinstance(item, guiitems.GuiItem), rectangledItems))
        for guiItem in filtered:
            try: #items like the MultiMeasureRest and Upbeat hide their data for the caching engine to work.
                api.remoteSelect(guiItem.dataOrg, guiItem.track.data)
            except:
                api.remoteSelect(guiItem.data, guiItem.track.data)

    def deselectInRectangle(self, qRectangleF):
        """Used in eventFilter for rubberband selection and
        mouseButtonPressed for Shift+Press"""
        #draw = QtGui.QGraphicsRectItem(qRectangleF)
        #self.scene.addItem(draw)
        rectangledItems = self.scene.items(qRectangleF, mode = QtCore.Qt.IntersectsItemShape)
        filtered = list(filter(lambda item: isinstance(item, guiitems.GuiItem), rectangledItems))
        for guiItem in filtered:
            api.remoteDeselect(guiItem.data, guiItem.track.data)

    def zoomIn(self):
        x = self.ui.score.transform().m11() #Y is the same
        newTransform = QtGui.QTransform()
        newTransform.scale(round(x+0.1, 2), round(x+0.1, 2))
        self.ui.score.setTransform(newTransform)
        #self.ui.score.centerOn(self.cursorPosition.pos())
        #The order of setIncipit Transform and cursor position is critical! I don't know why, but else zooming both views in sync does not work.
        newTransform2 = QtGui.QTransform()
        newTransform2.scale(1, round(x+0.1, 2))
        self.ui.incipit.setTransform(newTransform2)
        self.ui.score.viewport().update #it is possible that scrollbars have to appear. Update is needed for this.

    def zoomOut(self):
        x = self.ui.score.transform().m11() #Y is the same
        newTransform = QtGui.QTransform()
        newTransform.scale(round(x-0.1, 2), round(x-0.1, 2))
        self.ui.score.setTransform(newTransform)
        newTransform2 = QtGui.QTransform()
        newTransform2.scale(1, round(x-0.1, 2))
        self.ui.incipit.setTransform(newTransform2)
        #The order of setIncipit Transform and cursor position is critical! I don't know why, but else zooming both views in sync does not work.
        #It is exactly the opposite of zoomIn here.
        #self.ui.score.centerOn(self.cursorPosition.pos())
        self.ui.score.viewport().update #it is possible that scrollbars have to appear. Update is needed for this.

    def zoomNull(self):
        x = self.ui.score.transform().m11() #Y is the same
        newTransform = QtGui.QTransform()
        newTransform.scale(1,1)
        self.ui.score.setTransform(newTransform)
        self.ui.incipit.setTransform(newTransform)

    def _stretchXCoordinates(self, factor):
        config.DURATION_MOD *= factor
        for backendItem, item in self.allItems.items():
            for i in item:
                i.setX(i.pos().x() * factor) #adjust all track items
                if i.group: #adjust beam
                    i.updateGroupStemAndFlag()
        for tr in self.tracks:
            for barline in tr.barlines: #adjust barlines
                barline.setX(barline.pos().x() * factor)
            for trline in tr.lines: #adjust stafflines
                new = trline.line()
                new.setLength(new.length() * factor)
                trline.setLine(new)
        self.sendMaxTrackWidthToMain() #new scrollbars
        self.cursorPosition.update()

    def widen(self):
        self._stretchXCoordinates(1*1.2) #2 is good

    def shrinken(self):
        self._stretchXCoordinates(1/1.2) #0.5 is good

    def connectMenuAction(self, action, function, append = True):
        """Connect the menu action with the backend function,
        or qt function in a few cases like Save and Open.
        The actions already exist in their places and
        no shortcuts are getting set here."""
        def funcWrap():
            function()
            self.ui.score.viewport().update()

        self.connect(action, QtCore.SIGNAL("triggered()"), funcWrap)

    def generateMovementSubstitutionsMenu(self):
        """Look up all movement substitutions and generate a menu for
        them. This is self updating. Call it again if new substitutions
        are added to the movement."""
        mainMenu = main.placeholderMovementSubstitutions
        mainMenu.clear()

        if self.movementSubstitutionsMenu:
            self.movementSubstitutionsMenu.clear()
            self.movementSubstitutionsMenu = None

        self.movementSubstitutionsMenu = mainMenu

        alreadyProcessedGroups = {} #each group shall only create one submenu. No nested menus.

        for key, value in sorted(api._getMovement().substitutions.items()):  #key is the name string, value a backend Container.
            group = value.group
            if group and not group in alreadyProcessedGroups:
                newSubMenu = self.movementSubstitutionsMenu.addMenu(group)
                alreadyProcessedGroups[group] = newSubMenu
            pretty = " ".join(key.split("_")).title()
            if group:
                entry = alreadyProcessedGroups[group].addAction(key)
            else:
                entry = self.movementSubstitutionsMenu.addAction(key) #use the real name.
            entry.setText(pretty) #but display a pretty name.
            #self.connect(entry, QtCore.SIGNAL('triggered()'), lambda key=key: api.substitution(key))
            self.connectMenuAction(entry, lambda key=key: api.substitution(key))

    def generateMovementContainerMenu(self):
        """Look up all movement container and generate a menu for
        them. This is self updating. Call it again if new container
        are added to the movement."""

        mainMenu = main.placeholderMovementContainers
        mainMenu.clear()

        if self.movementContainerMenu:
            self.movementContainerMenu.clear()
            self.movementContainerMenu = None

        self.movementContainerMenu = mainMenu

        alreadyProcessedGroups = {} #each group shall only create one submenu. No nested menus.
        for key, value in api._getMovement().nonTrackContainers.items():  #key is the backend container, value is the instance count and not important.
            group = key.group
            if group and not group in alreadyProcessedGroups:
                newSubMenu = self.movementContainerMenu.addMenu(group)
                alreadyProcessedGroups[group] = newSubMenu
            pretty = " ".join(key.uniqueContainerName.split("_")).title()
            if group:
                entry = alreadyProcessedGroups[group].addAction(key.uniqueContainerName)
            else:
                entry = self.movementContainerMenu.addAction(key.uniqueContainerName)
            entry.setText(pretty) #but display a pretty name.
            self.connectMenuAction(entry, lambda key=key: api.insertContainer(key))
            #self.connect(entry,QtCore.SIGNAL('triggered()'), lambda key=key: api.insertContainer(key))

    def toggleBarlines(self, onOff):
        """Switch GUI barlines on or off.
        This alone is only a relay station and is not a switch in itself
        toggleBarlines in main decides what to do"""
        #switching on
        if onOff:
            for tr in self.tracks:
                tr.updateBarlines()
        #switching off
        else:
            for tr in self.tracks:
                tr.deleteBarlines()

    def changeGroup(self, groupText):
        """Changes the current active Track group and its color"""
        api.trackChangeAttribute("group", groupText)
        self.currentTrack.incipit.changeGroupColor()

    def changeNameMarker(self, nameText):
        """Changes the current active Track group and its color"""
        api.trackChangeAttribute("shortInstrumentName", nameText)
        self.currentTrack.incipit.changeNameMarker()



class WelcomeTab(QtGui.QWidget):
    """The first tab. Give direct control to mouse users"""
    def __init__(self):
        super(WelcomeTab, self).__init__()
        self.ui = Ui_Form()  #self.ui is the node to all qt widgets in this tab.
        self.ui.setupUi(self)

class GuiCursorPosition(QtGui.QGraphicsRectItem):
    def __init__(self, parent):
        super(GuiCursorPosition, self).__init__(0, 0,  3, 40)
        self.setOpacity(1)
        self.setBrush(QtGui.QColor("red"))
        self.pitch = api._getCursorPitch()
        self.setZValue(10)
        self.parent = parent # I belong to this parent movement/Workspace
        self.position = 0  #current X position, in pixel/guiTicks. Needed for the pitch cursor up and down without changing X.
        self.posX = 0 # a cache value.
        #self.highlightEffect = QtGui.QGraphicsColorizeEffect() #conveniently this is a unique effect. apply it one item and it deletes itself from all other items. But this is also non-convenient because it removes from all linked items.

    def update(self):
        #self.position = api._getFlatCursorIndex() #get the position directly from the backend
        #self.parent.currentTrack
        currentGuiItem = self.parent.currentTrack.getCurrentItem()
        posY = self.parent.currentTrack.pos().y()
        if currentGuiItem:
            self.posX = currentGuiItem.pos().x()
        else: #appending
            self.posX = self.parent.currentTrack.cachedTrackWidth * config.DURATION_MOD
        self.setPos(self.posX, posY)
        self.parent.cursorPitch.update()
        self.makeVisible()

    def makeVisible(self):
        """Check if the cursor is outside the visible working area. if not make it visible."""
        scenePos = self.parent.ui.score.mapFromScene(self.pos())
        cursorViewPosX = scenePos.x() #the cursor position in View coordinates
        cursorViewPosY = scenePos.y() #the cursor position in View coordinates. 0 is the view beginning,  has nothing to do with the geometry/main window boundaries.
        geo = self.parent.ui.score.geometry()

        if cursorViewPosY + TRACK_SIZE > geo.height (): #pagebreak down
            bar = self.parent.ui.score.verticalScrollBar()
            bar.setValue(self.pos().y() * self.parent.ui.score.transform().m11() - self.parent.ui.score.viewport().height() /2)
        elif cursorViewPosY < 10: #pagebreak down
            bar = self.parent.ui.score.verticalScrollBar()
            bar.setValue(bar.value() - self.parent.ui.score.viewport().height() /2 + 50)

        if cursorViewPosX + 50 >= geo.width(): #pagebreak forward
            bar = self.parent.ui.score.horizontalScrollBar()
            bar.setValue((self.posX-8) * self.parent.ui.score.transform().m11())
        elif cursorViewPosX < 10: #pagebreak backwards
            bar = self.parent.ui.score.horizontalScrollBar()
            bar.setValue(bar.value() - self.parent.ui.score.viewport().width() + 50)
            #last resort. The cursor is still not visible. We had a go to head etc.
            if self.parent.ui.score.mapFromScene(self.pos()).x() < 10:
                self.parent.ui.score.centerOn(self.posX, self.pos().y())


class GuiCursorPitch(QtGui.QGraphicsRectItem):
    def __init__(self, parent):
        super(GuiCursorPitch, self).__init__(5, 0,  20, 4)
        self.setOpacity(1)
        self.setBrush(QtGui.QColor("red"))
        self.pitch = api._getCursorPitch()
        self.setZValue(10)
        self.parent = parent # I belong to this parent movement/Workspace

    def update(self):
        x = self.parent.cursorPosition.posX
        y = self.parent.currentTrack.pos().y() + -1*(self.pitch / PITCH_MOD) + self.parent.currentTrack.cachedClefPitch
        self.setPos(x, y + 28)

class GuiMovement(MusicTab):
    """A gui movement holds all guiItems in plain python data types
    but it also the access point to the qgraphicsview and scene.
    The GuiMovement itself is also the Widget inside one Tab
    through inherited MusicTab"""

    def __init__(self, backendWorkspace):
        super(GuiMovement, self).__init__()
        self.tracks = [] #GuiTracks
        self.cursorPitch = GuiCursorPitch(self)
        self.scene.addItem(self.cursorPitch)
        self.cursorPosition = GuiCursorPosition(self)
        self.scene.addItem(self.cursorPosition)
        self.data = backendWorkspace
        self.fileName = None
        self.displacementMap = [] # A sorted list of pairs (tick-position, displacement) which contains the maximum displacement from all tracks. A very dynamic variable which gets changed and calculated all the time.
        def appendingItem():
            return [PSEUDO]
        self.allItems=defaultdict(appendingItem,[]) # All items. backend-instance as key, list of gui-items as value.  There is only one instance for each unique backend item. A linked item is twice the same. So we can actually use the instance as key in this dict to find all gui items based on that backend-item.

    def insertItemRelativeToContainerStart(self, backendItem, container, howManySteps, standaloneInsert):
        """This is a gui only update process. No backend data,
        especially not their cursors, are changed
        For each track with a containerStart item crawl from the
        beginning to the end (lefPart + rightPart) and stop on
        each containerStartItem, go right the howManySteps, insert
        the item. Continue to search right. Thank you, next track.

        In the end go back to the backend cursor position which did not
        change during the whole process.
        Then update the changes in the gui.

        This insert assumes that there are no recursive containers.

        We need to create new items from the backendItem here. One
        for each container instance.
        """
        itemType = type(backendItem)
        #typ = type(guiItem)
        #if typ is GuiMultiMeasureRest or typ is GuiUpbeat: #some GuiItems don't work with the backendData but are synced manually. Multimeasure Rests and Upbeats.
        #    data = guiItem.dataOrg
        #else:
        #    data = guiItem.data

        realSteps =  howManySteps -2 #-1 for 1vs0 counting. additional -1 because we are already at the container start.
        containerBackendStartItem = container[0] # the backend container start.
        allTracks = set()
        for tr in self.allItems[containerBackendStartItem]: #The list of guiItems connected to this backend Item. All container start in the current movement.
            allTracks.add(tr.track) #the guiTracks
        for tra in allTracks:
            tra.reset() #all items to track.rightPart
            #steps in nested containers do not count. The whole nested container counts as 1 step.
            countdown = realSteps
            while True:
                containerStack = [] #register to keep track of nested containers in the gui, where they are just items in a row, not real nested like in the backend.
                currentItem = tra._right() #returns the current item or false.
                if not currentItem:
                    break #Track End

                #print ("[InsertRelative]Found Match Start at", len(tra.rightPart), ". Items left:", len(tra.leftPart))  #len() returns counting from 1
                if currentItem.breakContainerLoop and currentItem.data == containerBackendStartItem: #we have found a containerStart and it is the same. Start a sub-loop.
                    while countdown > 0:
                        innerItem = tra.getCurrentItem()
                        if innerItem.breakContainerLoop: #container Start()!
                            containerStack.append(True)
                        elif innerItem.breakContainerLoop is None: #and not False! Container End()
                            #print (innerItem, innerItem.data, innerItem.data.exportLilypond())
                            containerStack.pop()

                        if not containerStack: #we are not in a nested container
                            countdown -= 1
                        tra._right()
                    else: #we reached 0 steps.
                        countdown = realSteps
                        before = tra.leftPart[-1]
                        #print (before.data.exportLilypon(), before.
                        forceSigs  = [tra, before.cachedClefPitch, before.cachedKeysigData, before.cachedTimesigData]
                        new = api.l_item._convertBackendItemToGuiItem(backendItem, itemType, forceSigs = forceSigs)  #the big typechecking machine
                        #new = typ(data) #create a new for each container. The gui only knows unique items. Essentialy that means we copy, and forget the original one.
                        #new = copy.copy(guiItem) #QGraphicsItems are NOT copy-able. Qt simply does not allow it.
                        if new.breakContainerLoop is None and tra.getCurrentItem().breakContainerLoop: #and not False! #a quick bugfix for container ends. Move one further. Else container start and end are reversed.
                            tra._right()
                        #Achtung: Inserting an empty container into a container end may result in two reversed containerEnd in the GUI. For the user it makes no difference. Only when we begin to utilize the containerEnd for something else, like drawing backgrounds, this will confuse the display.
                        tra.insert(new, standaloneInsert = standaloneInsert) #Insert at cursor position. Goes right after inserting.
                        self.allItems.setdefault(backendItem, []).append(new)

                #Continue looping, there may be more container in the same track. Actually that is very likely and the normal case.

        #finally update all necessary track durations and positioning.
        api.l_send(lambda: api.l_movement.updateTrackFromItem(backendItem)) #update cursor position is done in here also

    def deletetItemRelativeToContainerStart(self, backendItem, container, howManySteps):
        """Based on the insert variant"""
        realSteps =  howManySteps -1 #-1 for 1vs0 counting.
        containerBackendStartItem = container[0] # the backend container start.
        allTracks = set()
        for tr in self.allItems[containerBackendStartItem]: #The list of guiItems connected to this backend Item. All container start in the current movement.
            allTracks.add(tr.track) #the guiTracks
        for tra in allTracks:
            tra.reset() #all items to track.rightPart
            #steps in nested containers do not count. The whole nested container counts as 1 step.
            countdown = realSteps
            while True:
                containerStack = [] #register to keep track of nested containers in the gui, where they are just items in a row, not real nested like in the backend.
                currentItem = tra._right() #returns the current item or false.
                if not currentItem:
                    break #Track End

                if currentItem.breakContainerLoop and currentItem.data == containerBackendStartItem: #we have found a containerStart and it is the same. Start a sub-loop.
                    while countdown > 0:
                        innerItem = tra._right()
                        if innerItem.breakContainerLoop: #container start!
                            containerStack.append(True)
                        elif innerItem.breakContainerLoop is None: #and not False!
                            containerStack.pop()
                        if not containerStack: #we are not in a nested container
                            countdown -= 1
                    else: #we reached 0 steps
                        countdown = realSteps
                        tra.delete(backendItem)

                #Continue looping, there may be more container in the same track. Actually that is very likely and the normal case.

        #finally update all necessary track durations and positioning.
        api.l_send(lambda: api.l_movement.updateTrackFromItem(containerBackendStartItem)) #update cursor position is done in here also

    def loadBackendData(self):
        """Read in the backend movement and create GUI items.
        Used on start and load file.
        This is the only time where items are not inserted at
        the backendcursor position. So we have to make a few update runs
        """
        main.ui.centralwidget.hide()
        self.tracks = []
        for track in api._getMovement(): #First pass create all tracks. Otherwise api.trackDown() in the next step will trigger a gui track down which will not know where to go.
            self.addTrack(track)
        for track in api._getMovement():
            for item in track:
                api.l_item.insertAtCursor(item, getContainer =None, standaloneInsert = False) #don't update after every item insert. . #we are assuming a non-container based entry. This is safe on load (in opposite to cursor insert) because if there is a container it is already complete needs no checking if the cursor is in front of a ContainerStart etc.  insertAtCursor will take care of the recursive container insert and force getContainer to True
            api.trackDown()
        api.trackFirst()
        api.head()
        self.buildDisplacementMap()
        for track in self.tracks:
            track.signatureReset()
            track.updateRightFromStart()  #This creates the correct positioning. Until now all items were just added to 0,0 so that we don't get a gui update after each insert

        self.generateMovementSubstitutionsMenu()
        self.generateMovementContainerMenu()
        main.ui.centralwidget.show()

    def buildDisplacementMap(self):
        """Return a unified list for all tracks.
        Contains the maximum displacement for a tickposition
        from all tracks.
        It modifies an instance variable which is basically a chache."""
        combined_dicts = []
        for track in self.tracks:
            #Goal: [(384, 10), (384, 15), (768, 15)] => [(384,25), (768, 15)]
            d = defaultdict(int); #with a default dict we later can overwrite AND create a key with += instead of having to check for existence first.
            for guiItem in track.displacements:
                d[guiItem.cachedTickIndex] += guiItem.displacement
            combined_dicts.append(d)

        result = {}
        for d in combined_dicts:
            result.update((k, v) for k, v in d.items() if v > result.get(k, 0))
        self.displacementMap = sorted(result.items())

    def getDisplacement(self, tickindex, includeSelf):
        """Return the complete displacement up to a certain tickindex
        while you can choose to include the tickindex position itself
        or not"""
        complete = 0
        if includeSelf:
            for position, displacement in self.displacementMap:
                if position > tickindex:
                    break
                complete += displacement
        else:
            for position, displacement in self.displacementMap:
                if position == tickindex:
                    break
                complete += displacement

        return complete

    @property
    def currentTrackIndex(self):
        return api._getTrackIndex()

    @property
    def currentTrack(self):
        return self.tracks[self.currentTrackIndex]

    def swapTrackWithAbove(self):
        """Move the complete Track up which results in the upper Track moving down"""
        #When now is defined it is already the new situation because we derive tracknumber from the api
        above = self.currentTrackIndex
        now = above + 1
        self.tracks[now].setPos(0, above * TRACK_SIZE)
        self.tracks[above].setPos(0, now * TRACK_SIZE)
        self.tracks.insert(above , self.tracks.pop(now)) #swap the track
        self.repositionTracks()

    def swapTrackWithBelow(self):
        """Move the complete Track down which results in the lower Track moving up"""
        #When now is defined it is already the new situation because we derive tracknumber from the api
        below = self.currentTrackIndex
        now = below - 1
        self.tracks[now].setPos(0, below * TRACK_SIZE)
        self.tracks[below].setPos(0, now * TRACK_SIZE)
        self.tracks.insert(below , self.tracks.pop(now)) #swap the track
        self.repositionTracks()

    def addTrack(self, backendTrack):
        #Let's assume the backend tells us to create a track in the last position if api.addTrack() is called. So this is only for creating a gui track.
        track = GuiTrack(workspace = self, backendTrack = backendTrack)
        self.tracks.append(track)
        self.scene.addItem(track)
        self.incipitScene.addItem(track.incipit)
        self.repositionTracks()
        track.updateRight() #Do it once for the empty track. This shows the stafflines and costs nothing

    def deleteTrack(self, trackNumberToDelete, newBackendTrackIndex):
        """delete all the items and lines"""
        currentTrack = self.tracks[trackNumberToDelete]
        self.incipitScene.removeItem(currentTrack.incipit)
        self.scene.removeItem(currentTrack)
        del self.tracks[trackNumberToDelete]
        self.repositionTracks()
        if not self.tracks:  #was this the last track?
            self.addTrack(api._getTrack()) #there is only one left in the backend.

    def sendMaxTrackWidthToMain(self):
        self.graphicsViewRect[2] = max([trk.cachedTrackWidth for trk in self.tracks])  * config.DURATION_MOD
        #no need for incipitRectView updates here
        self.updateSceneRect()

    def repositionTracks(self):
        """Reposition all tracks according to their index.
        Very fast."""
        offsetsSoFar = 0
        trackPositions = [] # (highest, lowest, gui-track)
        oneLyricSpace = FONT_SIZE/1.5 if config.LYRICS else 0
        for indx, trk in enumerate(self.tracks):
            highestPos = indx * trk.size + offsetsSoFar/2
            trk.setPos(0, highestPos)
            trk.incipit.setPos(0, highestPos)
            trackPositions.append((highestPos, highestPos+trk.size+trk.offset+trk.cachedVerseCount*oneLyricSpace, trk))
            offsetsSoFar += (trk.offset + trk.cachedVerseCount*oneLyricSpace)
        self.graphicsViewRect[3] = len(self.tracks) * TRACK_SIZE
        #no need for incipitGraphicsViewRect update here.

        #The outer tracks have no boundaries.
        if len(trackPositions) >= 2:
            trackPositions[0] = (float("-inf"), trackPositions[0][1], trackPositions[0][2])
            trackPositions[-1] = (trackPositions[-1][0], float("inf"), trackPositions[-1][2])
        elif len(trackPositions) == 1:
            trackPositions = [(float("-inf"), float("inf"), self.tracks[0])]

        self.cachedTrackPositions = trackPositions #is used by the mouse click
        self.updateSceneRect()
        self.cursorPosition.update() #calls cursor pitch update

class Session(object):
    """All open files with their QGraphicScenes etc.
    Also manages the saveStatus warning. The savestatus system is purely
    a GUI one.
    If the saveStatus set is empty/clean there are no unsaved changes.
    So only the "save" function can call setSavestatusClean while
    a lot of functions can call it dirty.
    Load does not need any saveStatus access because initially loaded
    data is clean."""
    def __init__(self):
        self.scenes = {} # key is backend Workspace, value is the current workspace.
        self.saveStatus = set() #A set of all changed and unsaved backend workspaces. Movements check with "if self in saveStatus" and the program checks with "if saveStatus".
        #self.lastFileDialogState = None

    @property
    def currentGuiWorkspace(self):
        if api._getWorkspace() in self.scenes:
            return self.scenes[api._getWorkspace()]
        else:
            return None

    def isFilenameAvailable(self, newFileName):
        l = set()
        for backendMov, guiMov in self.scenes.items():
            l.add(guiMov.fileName)
        if newFileName in l:
            return False
        else:
            return True


    def isCurrentClean(self):
        """Check if the current active workspace/movement is clean"""
        return not api._getWorkspace() in self.saveStatus #return True if it is not in saveStatus, which means there is no unsaved Data which is good and therefore true.

    def setSavestatusClean(self):
        """Clean the save status for the current active workspace."""
        ws = api._getWorkspace()
        if ws in self.saveStatus:
            self.saveStatus.remove(ws)

    def setSavestatusDirty(self):
        """Add the current movement to the set of unclean data"""
        try:
            self.saveStatus.add(api._getWorkspace())
        except:
            pass

HOME = os.path.expanduser("~") #lin and win
